; Echo cyfrowe, konwersja A/C i C/A
; Rafa Baranowski (C) 2004

.INCLUDE "m16def.inc"		; SEGMENT DEKLARACJI

.EQU SYS_FREQ	= 16		; czstotliwo pracy w MHz
					; stae klawiatury matrycowej
.EQU K_KLAW		= DDRC	;   rejestr kierunku portu
.EQU O_KLAW		= PORTC	;   rejestr wyjciowy portu
.EQU I_KLAW		= PINC	;   rejestr wejciowy portu
.EQU K_KL_WCISN	= DDRD	;   rejestr kierunku linii "kl. wcinity"
.EQU O_KL_WCISN	= PORTD	;   rejestr wyjciowy linii "kl. wcinity"
.EQU I_KL_WCISN	= PIND	;   rejestr wejciowy linii "kl. wcinity"
.EQU KL_WCISN	= 2		;   nr linii dla sygnau "kl. wcinity"
					; stae modulacyjnego przetwornika C/A
.EQU K_PRZ_CA	= DDRB	;   rejestr kierunku portu
.EQU PRZ_CA		= 3		;   nr linii sygnaowej (OC0)

.DSEG					; SEGMENT DANYCH
.ORG	0x0060			; pocztek pamici SRAM
Klaw_inf:		.BYTE	1	; zmienna stanu klawiatury matrycowej
Bufor:		.BYTE (RAMEND-Bufor-15)	; bufor echa
Buf_kon:		.BYTE 1	; koniec bufora echa


.MACRO ldi_wsk			; makro adujce wskanik 16-bitowy sta 16-bitow
					; parametry:
					;   0. modyfikowany wskanik (X, Y lub Z)
					;   1. 16-bitowa bitowa warto bezporednia
	ldi	@0H, high(@1)
	ldi	@0L, low(@1)
.ENDM


.CSEG					; SEGMENT PROGRAMU
.ORG	0				; wektor zerowania
	jmp	Reset			; skok do procedury inicjalizujcej
.ORG	INT0addr			; wektor przerwania INT0
	jmp	Odczyt_klawiatury	; skok do procedury odczytujcej stan klawiatury
.ORG	ADCCaddr			; wektor przerwania ADCC
	reti				; procedura obsugi przerwania ADC jest pusta

.ORG	0x30				; przejcie do pierwszej komrki za wektorami przerwa


Reset:				; podprogram inicjalizujcy
	ldi_wsk X, RAMEND
	out	SPH, XH
	out	SPL, XL		; inicjalizacja wskanika stosu

	rcall	Ini_klawiatury	; inicjalizacja klawiatury matrycowej

	ldi	R16, (1<<REFS1)|(1<<REFS0)|(1<<ADLAR)
	out	ADMUX, R16		; przetwornik ADC pracuje z wewntrznym rdem
					;   napicia odniesienia (2,56V), a wynik konwersji
					;   jest przedstawiany w postaci przesunitej w lewo
	ldi	R16, (1<<ADEN)|(1<<ADIE)|(1<<ADPS2)|(1<<ADPS1)|(1<<ADPS0)|(1<<ADATE)
	out	ADCSR, R16		; przetwornik ADC pracuje w trybie wyzwalania
					;   automatycznego, przy taktowaniu f=fosc/128
					;   (125 kHz), przerwania dla ADC s aktywne
					;   (prbkowanie fp=fosc/128/13=9,6 kHz)
	in	R16, SFIOR
	cbr	R16, (1<<ADTS2)|(1<<ADTS1)|(1<<ADTS0)
	out	SFIOR, R16		; wyzwalanie przetwarzania ADC zachodzi po kadym
					;   zakoczeniu konwersji (tryb Free Run)
	in	R16, MCUCR
	cbr	R16, (1<<SM2)|(1<<SM1)|(1<<SM0)
	sbr	R16, (1<<SE)
	out	MCUCR, R16		; tryby upienia s dozwolone; trybem domylnym jest
					;   Idle (licznik 0. moe pracowa nieprzerwanie)
	ldi	R16, (1<<WGM01)|(1<<WGM00)|(1<<COM01)|(1<<CS00)
	out	TCCR0, R16		; licznik 0. pracuje w trybie szybkiej modulacji PWM,
					;   przy czstotliwoci fm=fosc=16 MHz (czstotliwo
					;   modulacji PWM wynosi zatem fm=fosc/256=62,5 kHz)

	sbi	K_PRZ_CA, PRZ_CA	; port modulatora PWM (linia OC0) pracuje jako wyjcie

	ldi_wsk X, Bufor		; inicjalizacja wskanika zapisu bufora echa
					;   (wskazuje on na pocztek bufora)
	ldi_wsk Y, Bufor+1	; inicjalizacja wskanika odczytu bufora echa
					;   (wyprzedza on o 1 wskanik zapisu)

	set				; ustawienie bitu T w SREG (pene echo)

	sei				; odblokowanie przerwa


Petla:				; ptla gwna programu
	lds	R16, Klaw_inf
	sbrs	R16, 7		; jeli nacinito dowolny klawisz, to..
	rjmp	P_spij

	cbr	R16, 1<<7		;   w wartoci klawisza zeruj bit informujcy o
					;     przytrzymaniu
	cpi	R16, 1		;   jeli nacinito klawisz "1", to..
	brne	P_kl_2
	ldi_wsk X, Bufor		;     ustaw wskaniki odczytu i zapisu bufora tak,
	ldi_wsk Y, Bufor+1	;       by efekt echa by wczony
	rjmp	P_spij		;     i przejd do trybu upienia

P_kl_2:
	cpi	R16, 2		;   jeli nacinito klawisz "2", to..
	brne	P_kl_4
	ldi_wsk X, Bufor		;     ustaw wskaniki odczytu i zapisu bufora tak,
	ldi_wsk Y, Bufor		;       by efekt echa by wyczony
	rjmp	P_spij		;     i przejd do trybu upienia

P_kl_4:
	cpi	R16, 4		;   jeli nacinito klawisz "4", to..
	brne	P_kl_5
	ldi	R16, (1<<ADEN)|(1<<ADIE)|(1<<ADPS2)|(1<<ADPS1)|(1<<ADPS0)|(1<<ADATE)
	out	ADCSR, R16		;     taktuj ADC czstotliwoci f=fosc/128
					;     (prbkowanie fp=fosc/128/13=9,6 kHz)
	rjmp	P_spij		;     i przejd do trybu upienia

P_kl_5:
	cpi	R16, 5		;   jeli nacinito klawisz "5", to..
	brne	P_kl_7
	ldi	R16, (1<<ADEN)|(1<<ADIE)|(1<<ADPS2)|(1<<ADPS1)|(1<<ADATE)
	out	ADCSR, R16		;     taktuj ADC czstotliwoci f=fosc/64
					;     (prbkowanie fp=fosc/64/13=19,2 kHz)
	rjmp	P_spij		;     i przejd do trybu upienia

P_kl_7:
	cpi	R16, 7		;   jeli nacinito klawisz "7", to..
	brne	P_kl_8
	set				;     wcz echo pene (ustaw T w SREG)
	rjmp	P_spij		;     i przejd do trybu upienia

P_kl_8:
	cpi	R16, 8		;   jeli nacinito klawisz "8", to..
	brne	P_spij
	clt				;     wcz echo niepene (zeruj T w SREG)


P_spij:
	sleep				; przejd w tryb upienia (z konwersj "w tle")

	in	R16, ADCH
	st	X+, R16		; zapisz w buforze wynik konwersji (prbk
					;   aktualn) i inkrementuj wskanik zapisu bufora
	ldi	R17, high(Buf_kon)
	cpi	XL, low(Buf_kon)
	cpc	XH, R17
	brne	P_wsk_X_ok		; jeli X wskazuje na koniec bufora echa,
	ldi_wsk X, Bufor		;   to  przywr go na pocztek
P_wsk_X_ok:

	ld	R17, Y+		; odczytaj z bufora prbk najstarsz
					;   (echo) i inkrementuj wskanik odczytu bufora
	subi	R17, 128		; odejmij od prbki echa skadow sta

	brts	P_pelne_echo	; jeli wczone jest echo niepene, to..
	asr	R17			;   podziel prbk echa przez 2

P_pelne_echo:
	subi	R16, 128		; odejmij od prbki aktualnej skadow sta
	add	R16, R17		; dodaj prbk echa do prbki aktualnej

	brvc	P_sygnal_ok		; jeli nastpio przepenienie formatu U2, to..

	brcs	P_niedomiar		;   jeli nastpi nadmiar, to..
	ldi	R16, 127		;     ustaw prbk aktualn na maksimum (127)
	rjmp	P_sygnal_ok
P_niedomiar:			;   jeli jednak wystpi niedomiar, to..
	ldi	R16, -128		;     ustaw prbk aktualn  na minimum (-128)

P_sygnal_ok:
	subi	R16, -128		; dodaj do prbki aktualnej skadow sta

	out	OCR0, R16		; i wg niej wyznacz wspczynnik wypenienia
					;    sygnau PWM na wyjciu OC0 (przetwarzanie C/A)
	ldi	R17, high(Buf_kon)
	cpi	YL, low(Buf_kon)
	cpc	YH, R17
	brne	P_wsk_Y_ok		; jeli Y wskazuje na koniec bufora echa,
	ldi_wsk Y, Bufor		;   to przywr go na pocztek

P_wsk_Y_ok:
	rjmp	Petla			; zaptl program


					; DOCZANE PODPROGRAMY
.INCLUDE "czekaj_ms.inc"	;   oczekiwanie milisekundowe
.INCLUDE "czekaj_us.inc"	;   oczekiwanie mikrosekundowe
.INCLUDE "klaw_mat.inc"		;   obsuga klawiatury matrycowej
